#include <windows.h>
#include <dinput.h>

// DirectInput stuff (have to be called after DD_Init)
extern "C" 
{
	#include "interface.h"
	#include "h_input.h"
// Kbd input
	unsigned char Keys[256], KeyPressed;
	DIJOYSTATE Joy;
}

// DirectInput stuff

LPDIRECTINPUT			pdi = NULL;
LPDIRECTINPUTDEVICE		pMouse = NULL;
LPDIRECTINPUTDEVICE		pKeyboard = NULL;
LPDIRECTINPUTDEVICE		pJoystick = NULL;

DIPROPRANGE diprg;

int JoyStickGood = NULL;

void ReleaseDInput()
{
	if (pKeyboard) {
		pKeyboard->Unacquire();
		pKeyboard->Release();
		pKeyboard = NULL;
	}
	if (pJoystick) {
		pJoystick->Unacquire();
		pJoystick->Release();
		pJoystick = NULL;
	}
	if (pdi) {
		pdi->Release();
		pdi = NULL;
	}	
}

// Init all DirectInput-devices (mouse + keyboard). 
BOOL InitDInput(HWND hWindow, HINSTANCE hInstance)
{
	HRESULT hr;

//** Already initialized
	if (pdi) return TRUE;

//** Init DirectInput
	hr = DirectInputCreate(hInstance, DIRECTINPUT_VERSION, &pdi, NULL);
	if (FAILED(hr)) 
		return FALSE;

//** Init keyboard
	hr = pdi->CreateDevice(GUID_SysKeyboard,&pKeyboard,NULL);
	if (FAILED(hr)) return FALSE;

	hr = pKeyboard->SetDataFormat(&c_dfDIKeyboard);
	if (FAILED(hr)) return FALSE;

	hr = pKeyboard->SetCooperativeLevel(hWindow, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
	if (FAILED(hr)) 
		return FALSE;

	hr = pdi->CreateDevice(GUID_Joystick,&pJoystick,NULL);
	if (FAILED(hr)) pJoystick=NULL;
	else
	{
		JoyStickGood = 1;
		hr = pJoystick->SetDataFormat(&c_dfDIJoystick);
		if (FAILED(hr)) JoyStickGood = 0;
		else
			hr = pJoystick->SetCooperativeLevel(hWindow, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
		if (FAILED(hr)) JoyStickGood = 0;
		else
		{
			diprg.diph.dwSize			= sizeof(diprg);
			diprg.diph.dwHeaderSize		= sizeof(diprg.diph);
			diprg.diph.dwObj			= DIJOFS_X;
			diprg.diph.dwHow 			= DIPH_BYOFFSET;
			diprg.lMax					=  127;
			diprg.lMin 					= -127;

			hr = pJoystick->SetProperty(DIPROP_RANGE, &diprg.diph);
		}
		if (FAILED(hr)) JoyStickGood = 0;

	}

//** acquire kbd
	pKeyboard->Acquire();
	if (pJoystick != NULL) pJoystick->Acquire();

	return TRUE;
}


//pJoystick

void DI_GetJoystickData(void)
{
	HRESULT hr;
	int done = 1;
	int cnt = 3;

	if (!pJoystick) return;

	while (done && cnt)
	{
		done = 0;
		cnt--;
		hr = pJoystick->GetDeviceState(sizeof(DIJOYSTATE), &Joy);
		if ((hr == DIERR_INPUTLOST) || (hr == DIERR_NOTACQUIRED ))
		{
			hr = pJoystick->Acquire();
			if (SUCCEEDED(hr)) done = 1;
		}
	}


	if (FAILED(hr))
	{
		JoyStickGood = 0;
	}
	if (SUCCEEDED(hr))
	{
		JoyStickGood = 1;
	}
}



void DI_GetKeyboardData(void)
{
	HRESULT hr;
	int done = 1;

	if (!pKeyboard) return;

	while (done)
	{
		done = 0;
		hr = pKeyboard->GetDeviceState(sizeof(Keys), Keys);
		if ((hr == DIERR_INPUTLOST) || (hr == DIERR_NOTACQUIRED ))
		{
			hr = pKeyboard->Acquire();
			if (SUCCEEDED(hr)) done = 1;
		}
	}


	if (SUCCEEDED(hr)) {
	}
}



extern "C" 
{
	void Input_GetData() 
	{
		DI_GetKeyboardData();
		if (JoyStickGood) DI_GetJoystickData();
	}

	BOOL InitDirectInput(HWND hWindow, HINSTANCE hInstance)
	{
		return (InitDInput(hWindow, hInstance));
	}

	void ReleaseDirectInput()
	{
		ReleaseDInput();
	}
	void DI_Unacquire(void)
	{
		if (pKeyboard) pKeyboard->Unacquire();
		if (pJoystick) pJoystick->Unacquire();
	}

	void DI_SyncAcquire(void)
	{
		if (!pKeyboard) return;
			pKeyboard->Acquire();
		if (!pJoystick) return;
			pJoystick->Acquire();
	}

	static WORD		   keystate = 0;
	static signed char stick_x = 0;
	static signed char stick_y = 0;
	DWORD GetDInputData()
	{
		int StickFlag = 0;
		Input_GetData();

	//** KeyPress and KeyReleases 

	//** Function-Buttons
		//** Start Button
		if (Keys[kQ])    
			keystate |= 0x1000;
		else
			keystate &= ~0x1000;

		//** A-Button
		if (Keys[kW])    
			keystate |= 0x8000;
		else
			keystate &= ~0x8000;

		//** B-Button
		if (Keys[kE])    
			keystate |= 0x4000;
		else
			keystate &= ~0x4000;

	//** C-Buttons...
		//** C-Button UP
		if (Keys[kHOME])    
			keystate |= 0x0008;
		else
			keystate &= ~0x0008;

		//** C-Button DOWN
		if (Keys[kEND])    
			keystate |= 0x0004;
		else
			keystate &= ~0x0004;

		//** C-Button LEFT
		if (Keys[kDEL])    
			keystate |= 0x0002;
		else
			keystate &= ~0x0002;

		//** C-Button RIGHT
		if (Keys[kPGDN])    
			keystate |= 0x0001;
		else
			keystate &= ~0x0001;

	//** D-Buttons...
		//** D-Button UP
		if (Keys[kI])    
			keystate |= 0x0800;
		else
			keystate &= ~0x0800;

		//** D-Button DOWN
		if (Keys[kK])    
			keystate |= 0x0400;
		else
			keystate &= ~0x0400;

		//** D-Button LEFT
		if (Keys[kJ])    
			keystate |= 0x0200;
		else
			keystate &= ~0x0200;

		//** D-Button RIGHT
		if (Keys[kL])    
			keystate |= 0x0100;
		else
			keystate &= ~0x0100;

	//** Special Keys
		//** L-Button 
		if (Keys[kA])    
			keystate |= 0x0020;
		else
			keystate &= ~0x0020;

	//** R-Button
		if (Keys[kS])    
			keystate |= 0x0010;
		else
			keystate &= ~0x0010;

	//** Z-Button
		if (Keys[kD])    
			keystate |= 0x2000;
		else
			keystate &= ~0x2000;

//************************************

	//** Stick UP
		if (Keys[kUP])
		{
			StickFlag = 1;
			if(stick_y < 0x60) stick_y += 0x20;
		} else
			if(stick_y > 0x0) stick_y -= 0x20;

	//** Stick DOWN
		if (Keys[kDOWN])
		{
			StickFlag = 1;
			if(stick_y > -0x60) stick_y -= 0x20;
		} else
			if(stick_y < 0x0) stick_y += 0x20; 
	//** Stick RIGHT  
		if (Keys[kRIGHT])
		{
			StickFlag = 1;
			if(stick_x < 0x60) stick_x += 0x20;
		} else
			if(stick_x > 0x0) stick_x -= 0x20;

	//** Stick LEFT
		if (Keys[kLEFT])
		{
			StickFlag = 1;
			if(stick_x > -0x60) stick_x -= 0x20;
		} else
			if(stick_x < 0x0) stick_x += 0x20;

	//** Release Stick
		if (StickFlag == 0)
		{
			stick_x = 0;
			stick_y = 0;
		}
		if (JoyStickGood)
		{
		//** Start Button
			if (Joy.rgbButtons[0] & 0x80) keystate |= 0x1000;

		//** A-Button
			if (Joy.rgbButtons[1] & 0x80) keystate |= 0x8000;

		//** B-Button
			if (Joy.rgbButtons[2] & 0x80) keystate |= 0x4000;

//			stick_x = Joy.lRx;
//			stick_y = Joy.lRy;
		}
		return(((DWORD)(keystate << 16)) | ((DWORD)(WORD)(stick_x << 8)) | ((DWORD)(byte)stick_y));
	}
} 