/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/
#include <stdio.h>
#include <string.h>

#include "../global/type_sizes.h"
#include "../global/memory_extern.h"

#include "../file_loader_extern.h"

#include "memory.h"
#include "interrupts.h"
#include "controller.h"
#include "registers.h"
#include "dma.h"
#include "rsp_registers.h"

//** dyna - start 
#include "dynaCompiler.h"
//** dyna - end 

/* the following routines should use real dma! */
/* ... but that is impossible under unix ... maybe we could use a thread? */
/* currently this works :-) */





/*
   write happened to:
   SP_WR_LEN_REG   (0x404000c)
*/

void dma_idmem2dram()
{
        #define DMEM 0x0000
        #define IMEM 0x1000

        unsigned int     i;

       uWORD    idmem_offset = (((uWORD *)mem.sp_reg)[0] & 0xfff);
        int     imem_or_dmem = (((uWORD *)mem.sp_reg)[0] & 0x1000);
        
       uWORD    dram_offset  = (((uWORD *)mem.sp_reg)[1] & 0xffffff);

       uWORD    dma_length   = (((uWORD *)mem.sp_reg)[3] & 0x00000fff);
//warning The following warnings are because 'count' and 'skip' are not implemented right now (Niki)
       uWORD    dma_count    = ((((uWORD *)mem.sp_reg)[3] & 0x000ff000) >> 12);   /* BUG: not impl */
       uWORD    dma_skip     = ((((uWORD *)mem.sp_reg)[3] & 0xfff00000) >> 20);   /* BUG: not impl */

        for(i=0; i<=dma_length; i++)
        {
                if(imem_or_dmem & IMEM)
                {
                        mem.rd_ram[(i+dram_offset)^3] = mem.sp_imem[(i+idmem_offset)^3];
                }
                else
                {
                        mem.rd_ram[(i+dram_offset)^3] = mem.sp_dmem[(i+idmem_offset)^3];
                }

        }

        reg.do_or_check_sthg |= RS4300I_DO_SP_INTERRUPT;



        #undef IMEM
        #undef DMEM
        
} /* void dma_idmem2dram() */





/*
   write happened to:
   SP_RD_LEN_REG   (0x4040008)
*/
void CheckDList();
void dma_dram2idmem()
{
        #define DMEM 0x0000
        #define IMEM 0x1000

        unsigned int     i;
        
       uWORD    idmem_offset = (((uWORD *)mem.sp_reg)[0] & 0xfff);
        int     imem_or_dmem = (((uWORD *)mem.sp_reg)[0] & 0x1000);
        
       uWORD    dram_offset  = (((uWORD *)mem.sp_reg)[1] & 0xffffff);

       uWORD    dma_length   = (((uWORD *)mem.sp_reg)[2] & 0x00000fff);
//warning The following warnings are because 'count' and 'skip' are not implemented right now (Niki)
       uWORD    dma_count    = ((((uWORD *)mem.sp_reg)[2] & 0x000ff000) >> 12);   /* BUG: not impl */
       uWORD    dma_skip     = ((((uWORD *)mem.sp_reg)[2] & 0xfff00000) >> 20);   /* BUG: not impl */


        for(i=0; i<=dma_length; i++)
        {
                if(imem_or_dmem & IMEM)
                {
                        mem.sp_imem[(i+idmem_offset)^3] = mem.rd_ram[(i+dram_offset)^3];
                }
                else
                {
                        mem.sp_dmem[(i+idmem_offset)^3] = mem.rd_ram[(i+dram_offset)^3];
						
                }
        }

        //putchar('\n');

        reg.do_or_check_sthg |= RS4300I_DO_SP_INTERRUPT;


    /**
    * The following checks if we have a graphics or audio task loaded.
    * This just should happen if the RSP is stopped.
    *
    * We also check if we have written to IMEM last time.
    * This might be a BUG!
    **/
        if(rsp_reg.halt == 1)
                if(imem_or_dmem & IMEM)
					CheckDList();

        #undef IMEM
        #undef DMEM
        
} /* void dma_dram2idmem() */






/*
   write happened to:
   SI_PIF_ADDR_RD64B_REG   (0x4600004)
*/

void dma_pif2dram()
{
        int   i;
       uWORD  dram_offset = ((uWORD *)mem.si_reg)[0];

        controller_check();
        
        for(i=0; i<64; i++)
        {
                mem.rd_ram[i+dram_offset] = mem.pi_ram[i];
        }

        reg.do_or_check_sthg |= RS4300I_DO_SI_INTERRUPT;


} /* void dma_pif2dram() */





/*
   write happened to:
   SI_PIF_ADDR_WR64B_REG   (0x4600010)
*/

void dma_dram2pif()
{
        int   i;
        uWORD  dram_offset = ((uWORD *)mem.si_reg)[0];

        for(i=0; i<64; i++)
        {
                mem.pi_ram[i] = mem.pi_ram_w[i] = mem.rd_ram[i+dram_offset];
        }

        //controller_check();

        reg.do_or_check_sthg |= RS4300I_DO_SI_INTERRUPT;

} /* void dma_dram2pif() */





void dma_pi_copy()
{
       unsigned int i;
        
       uWORD source = ((uWORD *)mem.pi_reg)[1];
       uWORD dest   = ((uWORD *)mem.pi_reg)[0];
       uWORD len    = ((uWORD *)mem.pi_reg)[3];

       
        /* BUG:? horrible hack */
        source &= 0x3ffffff;   /* 64MB ... assumed as max rom length */

        if((source+len) > (unsigned)rom.length)
                len = rom.length - source;
        
//        if((dest+len) > 0x0400000)
//                len = 0x0400000 - source;
             
        for(i=0; i<=len; i++)
        {
                mem.rd_ram[(i+dest)^3] = rom.image[(i+source)^3];
        }

		/* dyna - start */
		dynaInvalidate(dest,len+1);
		/* dyna - end */

        reg.do_or_check_sthg |= RS4300I_DO_PI_INTERRUPT;
        
} /* void dma_pi_copy() */


//** Check if DList
#define INVALID_TASK 0
#define GFX_TASK     1
#define AUDIO_TASK   2

int RDPTask;

void CheckDList()
{
        uWORD us, uds, dss;

        us  = ((uWORD *)mem.sp_dmem)[0xfd4>>2];  /* ucode size */
        uds = ((uWORD *)mem.sp_dmem)[0xfdc>>2];  /* ucode data size */
        dss = ((uWORD *)mem.sp_dmem)[0xfe4>>2];  /* dram stack size */

        switch(((uWORD *)mem.sp_dmem)[0xfc0>>2])
        {
            case 1:                             /* Graphic Task */
                RDPTask = GFX_TASK;
/*		        if(us != 0x1000)
                {
                        RDPTask = INVALID_TASK;
                }
                if(uds != 0x800)
                {
                        RDPTask = INVALID_TASK;
                }
                if(dss < 640)
                {
                        RDPTask = INVALID_TASK;
                }*/
                break;

            case 2:                             /* Audio Task */
                RDPTask = AUDIO_TASK;
/*              if(us > 0x1000)
                {
                        RDPTask = INVALID_TASK;
                }
                if(uds > 0x800)
                {
                        RDPTask = INVALID_TASK;
                }
                if(dss != 0)
                {
                        RDPTask = INVALID_TASK;
                }*/
                break;

            default:
				RDPTask = INVALID_TASK;
				break;
        } /* switch(mem.sp_dmem[0xfc0>>2]) */
} /* void rdp_dl_check() */
